.wy-nav-content {
    max-width: 1200rem !important;
    z-index: 999;
}

li p:last-child {
    margin-bottom: 12em !important;
    margin-top: 6em;
    font-size: 14px;
    line-height: 1.5;
    font-weight: 800 !important;
}

html {
    scroll-behavior: smooth;
}
