/* This file is managed by 'repo_helper'. Don't edit it directly. */

div.highlight {
	-moz-tab-size: 4;
	tab-size: 4;
}

.field-list dt, dl.simple dt {
	margin-top: 0.5rem;
}

div.versionchanged ul, div.versionremoved ul {
	margin-left: 20px;
	margin-top: 0;
}
